from django import template
from django.core.signing import Signer
from django.db.models import Model

from turbo import get_channel_name

register = template.Library()


@register.inclusion_tag("turbo/turbo_stream_source.html")
def turbo_stream_from(model_instance):
    # https://docs.djangoproject.com/en/3.1/topics/signing/
    signer = Signer()
    channel_name = get_channel_name(model_instance)
    signed_channel_name = signer.sign(channel_name)
    return {"signed_channel_name": signed_channel_name}


@register.simple_tag
def stream_id(target):
    if isinstance(target, Model):
        model_instance: Model = target
        model_name = model_instance._meta.verbose_name.lower()
        pk = model_instance.pk
        return f"{model_name}_{pk}"
    else:
        return f"{target.__str__().lower()}"



from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer
from django.apps import apps
from django.core.signing import Signer, BadSignature
from django.template.loader import render_to_string

from turbo import REMOVE

signer = Signer()


class TurboStreamException(Exception):
    pass


class TurboStreamsConsumer(JsonWebsocketConsumer):
    def connect(self):
        self.accept()

    def notify_model(self, event):
        model_label = event["model"]
        model = apps.get_model(model_label)
        app, model_name = model_label.lower().split(".")
        instance = model.objects.get(pk=event["pk"])
        event["context"].update(
            {
                "object": instance,
                model_name: instance,
            }
        )

        self.notify(event)

    def notify(
        self,
        event,
    ):
        extra_context = event["context"]
        action = event["action"]
        dom_target = event["dom_target"]
        template_context = {
            "action": action,
            "dom_target": dom_target,
        }
        # Remove actions don't have contents, so only add context for model
        # template if it's not a remove action.
        if action != REMOVE:
            template_context.update({"model_template": event.get("template")})
            template_context.update(extra_context)

        signed_channel_name = signer.sign(event["channel_name"])
        self.send_json(
            {
                "signed_channel_name": signed_channel_name,
                "data": render_to_string("turbo/stream.html", template_context),
            }
        )

    def receive_json(self, content, **kwargs):
        try:
            channel_name = signer.unsign(content["signed_channel_name"])
        except (BadSignature, KeyError):
            raise TurboStreamException(
                "Signature is invalid or not present. This could be due to a misbehaving client."
            )
        message_type = content["type"]
        if message_type == "subscribe":
            async_to_sync(self.channel_layer.group_add)(channel_name, self.channel_name)
        elif message_type == "unsubscribe":
            async_to_sync(self.channel_layer.group_discard)(
                channel_name, self.channel_name
            )



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.db.models import Model

default_app_config = "turbo.apps.TurboDjangoConfig"


def make_channel_name(model_label, pk):
    return f"BROADCAST-{model_label}-{pk}".lower()


def get_channel_name(identifier):
    if isinstance(identifier, Model):
        return _channel_name_for_instance(identifier)
    else:
        return identifier.__str__()


def _channel_name_for_instance(instance: Model):
    return make_channel_name(instance._meta.label, instance.pk)


# Model actions
CREATED = "CREATED"
UPDATED = "UPDATED"
DELETED = "DELETED"

# Turbo Streams CRUD operations
APPEND = "append"
PREPEND = "prepend"
REPLACE = "replace"
REMOVE = "remove"


def broadcast_stream(
    stream_target,
    dom_target,
    action,
    template,
    context,
    send_type="notify",
    extra_palyoad=None,
):
    """
    Send a Broadcast to all Websocket Clients registered to a specifc stream!
    """
    if extra_palyoad is None:
        extra_palyoad = dict()
    channel_layer = get_channel_layer()
    channel_name = get_channel_name(stream_target)
    async_to_sync(channel_layer.group_send)(
        channel_name,
        {
            "type": send_type,
            "action": action,
            "channel_name": channel_name,
            "template": template,
            "context": context,
            "dom_target": dom_target,
            **extra_palyoad,
        },
    )



from django.db.models import Model

import turbo
from turbo import (
    CREATED,
    UPDATED,
    DELETED,
    REPLACE,
    REMOVE,
    APPEND,
)


class BroadcastableMixin(object):
    broadcasts_to = []  # Foreign Key fieldnames to broadcast updates for.
    broadcast_self = (
        True  # Whether or not to broadcast updates on this model's own stream.
    )
    inserts_by = APPEND  # Whether to append or prepend when adding to a list (broadcasting to a foreign key).
    default_stream_action = APPEND
    turbo_streams_template = None

    def get_turbo_streams_template(self):
        if self.turbo_streams_template is not None:
            return self.turbo_streams_template
        app_name, model_name = self._meta.label.lower().split(".")
        return f"{app_name}/{model_name}.html"

    def get_action(self, model_action):
        if model_action == CREATED:
            streams_action = self.inserts_by
        elif model_action == UPDATED:
            streams_action = REPLACE
        elif model_action == DELETED:
            streams_action = REMOVE
        else:  # TODO: What should the default be?
            streams_action = REPLACE
        return streams_action

    def broadcast(self, model_action):
        streams_action = self.get_action(model_action)

        if self.broadcast_self:
            self.send_broadcast(self, streams_action)

        for field_name in self.broadcasts_to:
            if hasattr(self, field_name):
                self.send_broadcast(getattr(self, field_name), streams_action)
            else:
                self.send_broadcast(field_name, streams_action)

    def get_context(self):
        return dict()

    def send_broadcast(self, stream_target, stream_action):
        turbo.broadcast_stream(
            stream_target,
            self.get_dom_target(stream_target),
            stream_action,
            self.get_turbo_streams_template(),
            self.get_context(),
            send_type="notify.model",
            extra_palyoad={"pk": self.pk, "model": self._meta.model._meta.label},
        )

    def get_dom_target(self, target):
        if isinstance(target, Model):
            model: Model = target
            if self._meta.model != model:
                # Broadcast to self
                return self._meta.verbose_name_plural.lower()
            else:
                return f"{self._meta.verbose_name.lower()}_{self.pk}"
        else:
            return f"{target.lower()}"

    def save(self, *args, **kwargs):
        creating = self._state.adding
        super().save(*args, **kwargs)
        self.broadcast(CREATED if creating else UPDATED)

    def delete(self, *args, **kwargs):
        self.broadcast(DELETED)
        super().delete(*args, **kwargs)
